/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2012                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#ifndef TOPOLOGICAL_CUT_H
#define TOPOLOGICAL_CUT_H

#include <vcg/complex/complex.h>
#include <vcg/complex/algorithms/update/color.h>
#include <vcg/complex/algorithms/geodesic.h>
#include <vcg/space/index/grid_static_ptr.h>
#include <vcg/complex/algorithms/closest.h>

#include <algorithm>
namespace vcg {
namespace tri {

template <class MeshType>
class TopologicalCut
{
 public:
  typedef typename MeshType::VertexType VertexType;
  typedef typename MeshType::VertexPointer VertexPointer;
  typedef typename MeshType::FaceType FaceType;
  typedef typename MeshType::FacePointer FacePointer;
  typedef typename MeshType::EdgeType EdgeType;
  typedef typename MeshType::VertexIterator VertexIterator;
  typedef typename MeshType::EdgeIterator EdgeIterator;
  typedef typename MeshType::FaceIterator FaceIterator;
  typedef typename MeshType::ScalarType ScalarType;
  typedef GridStaticPtr <FaceType, ScalarType> TriMeshGrid;

class GeodesicPos
{
public:
  face::Pos<FaceType> p;
  float dist;

  bool operator < (const GeodesicPos & p) const
  {
    return dist > p.dist;
  }

  GeodesicPos(FaceType *f, int edgeInd)
  {
    p = face::Pos<FaceType>(f,edgeInd);
    face::Pos<FaceType> tp= p;
    tp.FlipF();
    tp.FlipE();
    tp.FlipV();
    dist = tp.V()->Q();
  }
};

/* Counts and returns the number of selected edges incident on a vertex vp;
 * it returns in ep the pointer to one of the selected edge (if any)
 */
static int CountSelectedEdges(VertexType *vp, EdgeType *&ep)
{
  ep=0;
  FaceType *fp = vp->VFp();
  face::Pos<FaceType> startPos(fp,vp);
  face::Pos<FaceType> curPos=startPos;
  int selNum=0;
  do
  {
    curPos.FlipE();
    curPos.FlipF();
    if(curPos.F()->FEp(curPos.E())->IsS())
    {
      ep = curPos.F()->FEp(curPos.E());
      selNum++;
    }
  }
  while(curPos!=startPos);
  return selNum;
}
/* Counts and returns the number of selected edges incident on a face fp;
 * If there are exactly two selected edges and both of them are not adjacent to a vertex with more than two incident selected edges
 * it returns in ep the pointer to only the un-selected edge.
 * Used to 'smooth' the chain of selected edges.
 */

static int CountSelectedEdges(FaceType *fp, int &eInd)
{
  int unselNum=0;
  for(int i=0;i<3;++i)
  {
    EdgeType *ep;
    if(CountSelectedEdges(fp->V(i),ep)>2)
      return -1;
    if(face::IsBorder(*fp,i)) return -1;

    if(!fp->FEp(i)->IsS())
    {
      unselNum++;
      eInd = i;
    }
  }
  return 3-unselNum;
}
// Aux Per-face class that we use when we cut the mesh.
// for each wedge of the face we keep the index of the vertex to be added
class NewVertInd
{
public:
  int v[3];
  NewVertInd(){v[0]=-1;v[1]=-1;v[2]=-1;}
};

class AvgPos
{
public:
  AvgPos() :sum(0,0,0),cnt(0) {}
  Point3<ScalarType> sum;
  int cnt;
};

static void SmoothAndProject(MeshType &curve, MeshType &surface, int iterNum)
{
  typename MeshType::template PerVertexAttributeHandle<AvgPos> AvgH =
      vcg::tri::Allocator<MeshType>:: template AddPerVertexAttribute<AvgPos> (curve,std::string("AvgPos"));

  TriMeshGrid grid;
  grid.Set(surface.face.begin(),surface.face.end());

  for(int i=0;i<iterNum;++i)
  {
    for(EdgeIterator ei=curve.edge.begin();ei!=curve.edge.end();++ei)
    {
      AvgH[ei->V(0)].sum += ei->V(1)->P();
      AvgH[ei->V(1)].sum += ei->V(0)->P();
      AvgH[ei->V(0)].cnt++;
      AvgH[ei->V(1)].cnt++;
    }
    for(VertexIterator vi=curve.vert.begin();vi!=curve.vert.end();++vi)
    {
      vi->P()= (vi->P()+AvgH[vi].sum) / float(AvgH[vi].cnt+1);
      AvgH[vi].sum=Point3<ScalarType>(0,0,0);
      AvgH[vi].cnt=0;
    }
    float maxDist = surface.bbox.Diag()/50;
    for(VertexIterator vi=curve.vert.begin();vi!=curve.vert.end();++vi)
    {
      float dist;
      Point3f closest;
      FacePointer fp = tri::GetClosestFaceBase(surface,grid,vi->P(),maxDist,dist,closest);
      if(fp!=NULL)
      {
        vi->P()=closest;
      }
    }

  }

  vcg::tri::Allocator<MeshType>::DeletePerVertexAttribute(curve,AvgH);
}

static bool SelectCuttingEdges(MeshType &m)
{
  Clean<MeshType>::RemoveUnreferencedVertex(m);
  Allocator<MeshType>::CompactVertexVector(m);

// Step 1)
  // Choose a face and store geodesic distance from the vertexes of this face to all the other vertexes of the mesh.
  std::vector<VertexPointer> seedVec;
  FacePointer startFace=&m.face[0];
  seedVec.push_back(startFace->V(0));
  seedVec.push_back(startFace->V(1));
  seedVec.push_back(startFace->V(2));
  tri::UpdateTopology<MeshType>::VertexFace(m);
  tri::Geodesic<MeshType>::Compute(m,seedVec);
  tri::UpdateColor<MeshType>::PerVertexQualityRamp(m);
  tri::UpdateTopology<MeshType>::FaceFace(m);
  tri::UpdateTopology<MeshType>::AllocateEdge(m);

  // Step 1)
  // Make a topologic visit of the mesh in distance order.
  // Use a priority queue ordered according to the previously computed geodesic distance.
  // Once two faces expand onto a common edge select this edge;
  // In this way you build a cutting tree that unfold the mesh into a disk

  std::vector<GeodesicPos> heap;
  heap.push_back(GeodesicPos(startFace,0));
  heap.push_back(GeodesicPos(startFace,1));
  heap.push_back(GeodesicPos(startFace,2));

  make_heap(heap.begin(),heap.end());

  tri::UpdateSelection<MeshType>::FaceClear(m);
  startFace->SetS();
  int cnt=0;
  while(!heap.empty())
  {
    pop_heap(heap.begin(),heap.end());
    GeodesicPos cur = heap.back();
    heap.pop_back();
    FaceType *curFace = cur.p.FFlip();
    if(!curFace->IsS())
    {
      curFace->SetS();
      for(int i=0;i<3;++i)
      {
        if(!curFace->FFp(i)->IsS())
          heap.push_back(GeodesicPos(curFace,i));
        if(face::IsBorder(*curFace,i))
          curFace->FEp(i)->SetS();
      }
      make_heap(heap.begin(),heap.end());
    }
    else
    {
      EdgeType *ep = cur.p.F()->FEp(cur.p.E());
      ep->SetS();
    }
  }

  // Step 3)
  // Prune the tree, removing all dangling edges and leaving only the non-tree component (loops)
  // We use a stack of active vertices and deselect all dangling edges

  cnt=0;
  std::vector<VertexType *> pruneStack;
  for(VertexIterator vi=m.vert.begin();vi!=m.vert.end();vi++)
    pruneStack.push_back(&*vi);

  while(!pruneStack.empty())
  {
    VertexType *vp = pruneStack.back();
    pruneStack.pop_back();
    EdgeType *ep=0;
    if(CountSelectedEdges(vp,ep)==1)
    {
      assert(ep!=0);
      ep->ClearS();
      if(ep->V(0)==vp) pruneStack.push_back(ep->V(1));
      else pruneStack.push_back(ep->V(0));
    }
  }
}

static bool Mesh(MeshType &m)
{
  SelectCuttingEdges(m);

  // Step 4)
  // Trivial path smoothing
  // Shortcut each pair of edges as below
  //        /\
  //       /  \      --INTO-->
  //      /    \
  // ___./      \._____    _____.________.______
  //
  std::vector<FaceType *> zigzagStack;
  for(FaceIterator fi=m.face.begin();fi!=m.face.end();fi++)
    zigzagStack.push_back(&*fi);
  while(!zigzagStack.empty())
  {
    FaceType *fp = zigzagStack.back();
    zigzagStack.pop_back();
    int eind=-1;
    if(CountSelectedEdges(fp,eind)==2)
    {
      fp->FEp(eind)->SetS();
      fp->FEp((eind+1)%3)->ClearS();
      fp->FEp((eind+2)%3)->ClearS();
      zigzagStack.push_back(fp->FFp(eind));
    }
  }

  // Step 5)
  // Actually Cutting the mesh
  // We start from a border and we walk over it remainng always on the same side.
  // During the walk we add to all the wedges of the touched faces the index of a new vertex
  // At the end of the walk we actually add vertices and substitute the vertex references
  // in all the faces using the indexes that we have previously computed


  tri::UpdateFlags<MeshType>::VertexClearV(m);
  face::Pos<FaceType> startPos;
  // Find a first face with a selected edge
  for(FaceIterator fi=m.face.begin();fi!=m.face.end();fi++)
  {
    for(int i=0;i<3;++i)
      if(fi->FEp(i)->IsS())
      {
        startPos =  face::Pos<FaceType>(&*fi,i,fi->V(i));
         assert(startPos.F()->FEp(startPos.E())->IsS());
      }
  }

  typename  MeshType::template PerFaceAttributeHandle <NewVertInd>  nvH
      = tri::Allocator<MeshType>::template AddPerFaceAttribute < NewVertInd > (m);

  // Walk along the border...
  int nvi =0;
  face::Pos<FaceType> curPos=startPos;
  do
  {
    assert(curPos.F()->FEp(curPos.E())->IsS());
    curPos.V()->SetV();
    do
    {
      nvH[curPos.F()].v[curPos.VInd()] = nvi;
      curPos.FlipE();
      curPos.FlipF();
    }
    while (!curPos.F()->FEp(curPos.E())->IsS());

    curPos.FlipF();
    curPos.FlipV();
    nvi++;
  }
  while(curPos!=startPos);
  // After the walk we actually add the vertexes and substitute the vertex ref in the faces.
  VertexIterator vi=tri::Allocator<MeshType>::AddVertices(m,nvi);
  for(FaceIterator fi=m.face.begin();fi!=m.face.end();fi++)
  {
    for(int i=0;i<3;++i)
    {
      if(nvH[fi].v[i]>=0) // we touch only the vertex ref that we have previously filled
      {
        (vi+nvH[fi].v[i])->P() = fi->V(i)->P();
        fi->V(i)=&*(vi+nvH[fi].v[i]);
      }
    }
  }

  return 0;
}
};

} // end namespace
} // end namespace

#endif // TOPOLOGICAL_CUT_H
